@@ -1,6 +1,6 @@
 using DecFP, Test, Printf, Random, Base.MathConstants, SpecialFunctions
 
-@test DecFP.flags[Threads.threadid()] == 0
+include("printf.jl")
 
 import DecFP.isnanstr
 @test isnanstr("nan") && isnanstr("  +NAN") && isnanstr("-NaN") && !isnanstr("nano")
@@ -142,47 +142,7 @@ for T in (Dec32, Dec64, Dec128)
     show(io, T("123.456")); @test String(take!(io)) == "123.456"
     show(io, T("0.00123456")); @test String(take!(io)) == "0.00123456"
 
-    @test @sprintf("%7.2f", T("1.2345")) == "   1.23"
-    @test @sprintf("%-7.2f", T("1.2345")) == "1.23   "
-    @test @sprintf("%07.2f", T("1.2345")) == "0001.23"
-    @test @sprintf("%.0f", T("1.2345")) == "1"
-    @test @sprintf("%#.0f", T("1.2345")) == "1."
-    @test @sprintf("%.4e", T("1.2345")) == "1.2345e+00"
-    @test @sprintf("%.4E", T("1.2345")) == "1.2345E+00"
-
-    @test @sprintf("%6.2f", T("9.999")) == " 10.00"
-    @test @sprintf("%9.2e", T("9.999e5")) == " 1.00e+06"
-
-    @test @sprintf("%f", T("Inf")) == "Inf"
-    @test @sprintf("%f", T("NaN")) == "NaN"
-
-    @test @sprintf("%.0e", T("3e42")) == "3e+42"
-    @test @sprintf("%#.0e", T("3e42")) == "3.e+42"
-
-    @test @sprintf("%e", T("3e42")) == "3.000000e+42"
-    @test @sprintf("%E", T("3e42")) == "3.000000E+42"
-    @test @sprintf("%e", T("3e-42")) == "3.000000e-42"
-    @test @sprintf("%E", T("3e-42")) == "3.000000E-42"
-
-    @sprintf("%.6g", T("12345670.")) == "1.23457e+07"
-    @sprintf("%.6g", T("1234567.")) == "1.23457e+06"
-    @sprintf("%.6g", T("123456.7")) == "123457"
-    @sprintf("%.6g", T("12345.67")) == "12345.7"
-    @sprintf("%.6g", T("12340000.0")) == "1.234e+07"
-
-    @test @sprintf("%10.5g", T("123.4")) == "     123.4"
-    @test @sprintf("%+10.5g", T("123.4")) == "    +123.4"
-    @test @sprintf("% 10.5g", T("123.4")) == "     123.4"
-    @test @sprintf("%#10.5g", T("123.4")) == "    123.40"
-    @test @sprintf("%-10.5g", T("123.4")) == "123.4     "
-    @test @sprintf("%-+10.5g", T("123.4")) == "+123.4    "
-    @test @sprintf("%010.5g", T("123.4")) == "00000123.4"
-    @test @sprintf("%10.5g", T("-123.4")) == "    -123.4"
-    @test @sprintf("%010.5g", T("-123.4")) == "-0000123.4"
-    @test @sprintf("%.6g", T("12340000.0")) == "1.234e+07"
-    @test @sprintf("%#.6g", T("12340000.0")) == "1.23400e+07"
-
-    @test @sprintf("%.2f %.4f", T("12.34567"), 9.87) == "12.35 9.8700"
+    printftests(T)
 
     x,y,z = 1.5, -3.25, 0.0625 # exactly represented in binary
     xd = T(x); yd = T(y); zd = T(z)
@@ -464,8 +424,6 @@ for T in (Dec32, Dec64, Dec128)
     @test read(IOBuffer(take!(io)), T) === one(T)
 end
 
-@test DecFP.flags[Threads.threadid()] == 0
-
 @test widen(Dec32) == Dec64
 @test widen(Dec64) == Dec128
 @test widen(Dec32(1)) === Dec64(1)
